package com.me.asteroids;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.me.asteroids.components.AccelerationComponent;
import com.me.asteroids.components.AsteroidComponent;
import com.me.asteroids.components.BulletComponent;
import com.me.asteroids.components.ColliderComponent;
import com.me.asteroids.components.DebrisComponent;
import com.me.asteroids.components.DecayComponent;
import com.me.asteroids.components.ModelComponent;
import com.me.asteroids.components.PlayerComponent;
import com.me.asteroids.components.PositionComponent;
import com.me.asteroids.components.VelocityComponent;
import com.me.asteroids.components.model.LineModel;
import com.me.asteroids.components.model.Model;
import com.me.asteroids.components.model.PolygonModel;
import com.me.common.ecs.Engine;
import com.me.common.ecs.Entity;

import java.util.Arrays;

import static com.me.asteroids.Components.ASTEROID;
import static com.me.asteroids.Components.MODEL;
import static com.me.asteroids.Components.POSITION;
import static com.me.asteroids.Components.VELOCITY;
import static com.me.asteroids.Constants.rand;

public class EntityFactory {

    private static final Vector2 tmp = new Vector2();
    private static final Vector2 tmp2 = new Vector2();
    private static final Vector2 tmpA = new Vector2();
    private static final Vector2 tmpB = new Vector2();

    private static final ColliderComponent COLLIDER = new ColliderComponent();
    private static final DebrisComponent DEBRIS = new DebrisComponent();

    private static Entity createEntity(Engine engine) {
        Entity entity = engine.createEntity();
        if (Constants.DEBUG) {
            System.out.println("Spawned entity: " + entity);
        }
        return entity;
    }

    public static Entity createPlayer(Engine engine) {
        VelocityComponent velocity = new VelocityComponent(0f, 0f, 0f, 10f);

        ModelComponent playerModel = new ModelComponent(new PolygonModel(Color.WHITE));
        playerModel.setVertices(new float[]{
                0f, 0.5f,         // tip
                -5/16f, -0.5f,     // bottom left
                -1/8f, -5/16f,     // indent
                1/8f, -5/16f,      // indent
                5/16f, -0.5f,      // bottom right
        });

        ModelComponent afterburnerModel = new ModelComponent(new LineModel(Color.BLUE));
        afterburnerModel.setVertices(new float[]{
                -2/16f, -5/16f,
                0f, -0.8f,
                2/16f, -5/16f
        });

        Entity player = createEntity(engine);
        Entity afterBurner = createEntity(engine);

        player.addComponent(new PositionComponent(Constants.HALF_WIDTH, Constants.HALF_HEIGHT, 90));
        player.addComponent(velocity);
        player.addComponent(playerModel);
        player.addComponent(new AccelerationComponent(0f, 1f));
        player.addComponent(COLLIDER);
        player.addComponent(new PlayerComponent(afterBurner));
        player.setTag("PLAYER");

        afterBurner.addComponent(new PositionComponent(Constants.HALF_WIDTH, Constants.HALF_HEIGHT, 90));
        afterBurner.addComponent(velocity);
        afterBurner.addComponent(afterburnerModel);

        return player;
    }


    private static LineModel[] getLineModels(Model playerModel) {
        float[] vertices = playerModel.getVertices();

        LineModel[] models = new LineModel[(vertices.length / 2)];
        for (int aIndex = 0, n = vertices.length; aIndex < n; aIndex += 2) {
            int bIndex = (aIndex + 2 == n) ? 0 : (aIndex + 2);

            tmpA.set(vertices[aIndex], vertices[aIndex + 1]);
            tmpB.set(vertices[bIndex], vertices[bIndex + 1]);

            tmp.set(tmpB).sub(tmpA).scl(0.5f);
            tmp2.set(tmpA).add(tmp);                // tmp2 = center of line segment

            tmpA.sub(tmp2);                         // make tmpA, tmpB relative to line center
            tmpB.sub(tmp2);

            LineModel model = new LineModel(playerModel.getColor());
            model.setVertices(new float[]{
                    tmpA.x, tmpA.y,
                    tmpB.x, tmpB.y,
            });
            model.setPosition(tmp2);                // Position = line center

            models[aIndex / 2] = model;
        }
        return models;
    }

    public static Entity[] createDebris(Engine engine, Entity entity) {
        Vector2 playerVelocity = VELOCITY.get(entity).velocity;
        PositionComponent playerPosition = POSITION.get(entity);
        LineModel[] models = getLineModels(MODEL.get(entity).model);

        Vector2 explosionCenter = tmp.set(playerPosition.position).sub(Utils.setUnitVectorAngle(tmp2, playerPosition.rotation).scl(0.125f));

        Entity[] entities = new Entity[models.length];
        for (int i = 0, n = models.length; i < n; i++) {
            ModelComponent model = new ModelComponent(models[i]);

            VelocityComponent velocity = new VelocityComponent();
            velocity.set(new Vector2(models[i].getPosition())
                    .sub(explosionCenter)
                    .nor()                                      // Direction from explosion center to center of piece
                    .rotate(rand.nextFloat(-15, 15))            // Slightly alter the direction each piece flies off in
                    .scl(rand.nextFloat(2f, 2.5f))              // Give each piece a slightly different speed
                    .add(tmp2.set(playerVelocity).scl(0.75f))); // Maintain 75% of the player's velocity at impact
            velocity.angularVelocity = rand.nextFloat(-60, 60); // Make each piece spin at a different rate

            Entity debris = createEntity(engine);
            debris.addComponent(new PositionComponent(new Vector2(model.getPosition()), 90));
            debris.addComponent(velocity);
            debris.addComponent(model);
            debris.addComponent(DEBRIS);
            debris.addComponent(new DecayComponent(rand.nextFloat(0.5f, 2.5f)));
            entities[i] = debris;
        }
        return entities;
    }


    public static Entity createBullet(Engine engine, Entity player) {
        float[] modelVertices = player.getComponent(ModelComponent.class).model.getVertices();
        float rotation = player.getComponent(PositionComponent.class).rotation;

        PositionComponent position = new PositionComponent(modelVertices[0], modelVertices[1], rotation);
        ModelComponent model = new ModelComponent(new PolygonModel(Color.YELLOW));
        model.setVertices(new float[]{
                1/40f, 0f,
                -1/40f, 0f,
                -1/40f, -4/40f,
                1/40f, -4/40f,
        });
        model.setRotation(position.rotation);
        model.setPosition(position.position);

        Vector2 direction = Utils.setUnitVectorAngle(tmp, rotation);

        Entity bullet = createEntity(engine);
        bullet.addComponent(position);
        bullet.addComponent(new VelocityComponent(new Vector2(direction).scl(12.5f)));
        bullet.addComponent(model);
        bullet.addComponent(COLLIDER);
        bullet.addComponent(new BulletComponent());
        return bullet;
    }

    public static Entity createAsteroid(Engine engine) {
        // Creates an asteroid entity with position and velocity unset - the AsteroidSpawningSystem
        // is responsible for setting its position and velocity
        ModelComponent model = new ModelComponent(new PolygonModel(Color.WHITE));
        float size = rand.nextFloat(1f, 1.75f);
        model.setVertices(new AsteroidFactory()
                .setVertexCount(32)
                .setSize(size)
                .setSizeVariation(size * 0.5f)
                .sizeRelativeToLast()
                .generate());

        Entity asteroid = createEntity(engine);
        asteroid.addComponent(new PositionComponent());
        asteroid.addComponent(new VelocityComponent());
        asteroid.addComponent(model);
        asteroid.addComponent(COLLIDER);
        asteroid.addComponent(new AsteroidComponent());
        asteroid.addGroup("ASTEROIDS");
        return asteroid;
    }

    private static float[] scaleAndRelativizeVertices(Vector2 position, float[] vertices, float scale) {
        float[] ret = new float[vertices.length];
        for (int j = 0, n = vertices.length; j < n; j += 2) {
            ret[j] = (vertices[j] - position.x) * scale;
            ret[j + 1] = (vertices[j + 1] - position.y) * scale;
        }
        return ret;
    }

    public static Entity[] splitAsteroidIntoChunks(Engine engine, Entity asteroid, int chunkCount, float chunkScale) {
        Vector2 asteroidVelocity = VELOCITY.get(asteroid).velocity;
        ModelComponent asteroidModel = MODEL.get(asteroid);
        Vector2 asteroidPosition = asteroidModel.getPosition();
        float[] scaledVertices = scaleAndRelativizeVertices(asteroidPosition, asteroidModel.getVertices(), chunkScale);

        float angle = rand.nextFloat() * MathUtils.PI2;
        float angleStep = MathUtils.PI2 / chunkCount;
        Utils.setUnitVectorAngleRad(tmp, angle);

        Entity[] entities = new Entity[chunkCount];
        for (int i = 0; i < chunkCount; i++) {
            Vector2 chunkPosition = tmp2.set(asteroidPosition).add(tmp.scl(0.5f));

            ModelComponent model = new ModelComponent(new PolygonModel(asteroidModel.getColor()));
            model.setVertices(Arrays.copyOf(scaledVertices, scaledVertices.length));
            model.setPosition(chunkPosition);

            VelocityComponent velocity = new VelocityComponent();
            velocity.set(tmp);
            velocity.velocity.nor().rotate(rand.nextFloat(-45, 45)).scl(asteroidVelocity.len() * 1.10f);
            velocity.angularVelocity = rand.nextFloat(-30, 30);

            angle += angleStep;
            Utils.setUnitVectorAngleRad(tmp, angle);

            Entity split = createEntity(engine);
            split.addComponent(model);
            split.addComponent(new PositionComponent(new Vector2(chunkPosition), 90));
            split.addComponent(velocity);
            split.addComponent(new AsteroidComponent(ASTEROID.get(asteroid).generation + 1));
            split.addComponent(COLLIDER);
            split.addGroup("ASTEROIDS");
            entities[i] = split;
        }

        return entities;
    }

}
