package com.me.common.ecs;

import com.me.common.ecs.event.Event;
import com.me.common.ecs.event.EventHandler;
import com.me.common.ecs.event.Listener;
import com.me.common.ecs.event.RegisteredListener;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

class ListenerRegistry {

    private final Map<Class<? extends Event>, List<RegisteredListener>> registeredListeners;

    ListenerRegistry() {
        this.registeredListeners = new HashMap<>();
    }

    void registerListener(Listener listener) {
        for (Method method : listener.getClass().getMethods()) {
            EventHandler eh = method.getAnnotation(EventHandler.class);
            if (eh == null) {
                continue;
            }

            Class<?>[] parameters = method.getParameterTypes();
            if (parameters.length != 1 || !Event.class.isAssignableFrom(parameters[0])) {
                System.err.println("Attempting to register an EventHandler with an invalid method signature: " + method.toGenericString() + " in " + listener.getClass());
                continue;
            }
            method.setAccessible(true);

            Class<? extends Event> eventClass = parameters[0].asSubclass(Event.class);
            List<RegisteredListener> executors = getRegisteredListeners(eventClass);
            executors.add(new RegisteredListener(listener, method, eventClass, eh));
        }
    }

    void callEvent(Event event) {
        List<RegisteredListener> listeners = getRegisteredListeners(event.getClass());
        if (listeners.isEmpty()) {
            return;
        }

        for (RegisteredListener listener : listeners) {
            if (listener.isIgnoringCancelled() && event.isCancelled()) {
                continue;
            }
            listener.execute(event);
        }
    }

    private List<RegisteredListener> getRegisteredListeners(Class<? extends Event> eventClass) {
        List<RegisteredListener> listeners = registeredListeners.get(eventClass);
        if (listeners == null) {
            listeners = new ArrayList<>();
            registeredListeners.put(eventClass, listeners);
        }
        return listeners;
    }

}
